﻿#pragma once

#include <eigen3/Eigen/Dense>
#include "kalman_filter.h"
#include <vector>
#include <memory>

#define _PI_ 3.1415926


class BaseTrack 
{
public:
    // Constructor
    BaseTrack(unsigned int num_states, unsigned int num_obs);

    // Destructor
    ~BaseTrack() = default;

    //virtual std::shared_ptr<mytracker::KalmanFilter> InitKF(unsigned int num_states, unsigned int num_obs) { return nullptr; };
    virtual void Init(const std::vector<float>& data);
    virtual void Predict();
    virtual void Update(const std::vector<float>& data);
    virtual void UpdateDataCheck(const std::vector<float>& data, std::vector<float>& out);//对于输入数据进行修正
    virtual void UpdateHit();
    virtual int GetStateData(std::vector<float>& data);
    virtual int GetPredictData(std::vector<float>& data);
    virtual float GetNIS() const;
    virtual float GetProb() const;
    virtual bool IsLost();//数据是否丢失，如果不更新就会丢失
    virtual bool IsValid();//数据是否有效

    virtual int GetIouData(std::vector<float>& data,int& obj_type);
    
    virtual int GetIouDataOrder(std::vector<int>& order) = 0;
    virtual int GetKFDataOrder(std::vector<int>& order) = 0;

    virtual double CalculateIou(const std::vector<float>& data) = 0;

    int GetStatesNum();
    int GetObsNum();

    float* GetStatesXPtr();
    float* GetPredictPtr();
    float* GetRPtr();

    void SetIouThreshold(float threshold) { m_iou_threshold = threshold; }
    void SetMaxCoastCycles(int cycles) { m_kMaxCoastCycles = cycles; }
    void SetValidUpdateCount(int count) { m_updateValidCount = count; }
    virtual void SetValues(std::vector<float>& data) {}

    int coast_cycles_ = 0, hit_streak_ = 0;

    int m_update_count = 0;//数据更新的次数

    int m_num_states = 0;
    int m_num_obs = 0;

    float m_prob = 0.0f;//计算iou的配置度

    float m_iou_threshold = 0.01;//iou计算之后匹配的最小值
    int m_kMaxCoastCycles = 2;//predict之后，计算几次认为丢失。
    int m_updateValidCount = 3;//更新多少次认为是一个有效数据。


    std::shared_ptr<mytracker::KalmanFilter> kf_ = nullptr;
};
