﻿#pragma once


#include <iostream>
//#include <eigen3/Eigen/Dense>
//#include "kalman_filter.h"
#include <map>
//#include "BaseTrack.h"
#include <memory>
#include <vector>
#include "Iou.h"
//#include "LogBase.h"
#include <memory.h>
#include "Component.h"
#ifdef _KF_IOU_CUDA_
#include "bev_overlap_online.h"
#include "kalman_update_batch_online.h"
#endif

#ifdef _USING_NSIGHT_
#include <nvToolsExt.h>
#endif

template<class T>
class BaseTracker
{
public:
	BaseTracker() {}

    void SetGPU(int gpu) { m_isGPU = gpu; }
    void SetIouThreshold(float threshold) { m_iou_threshold = threshold; }
    void SetMaxCoastCycles(int cycles) { m_kMaxCoastCycles = cycles; }
    void SetValidUpdateCount(int count){ m_updateValidCount = count; }
    void SetValues(std::vector<float>& values) { m_values = values; }

    int Run(const std::vector<std::vector<float> >& detections, int _no/*观测数量*/, int _ns/*状态数量*/, std::vector<uint64_t>& detectionsId, std::map<uint64_t, int>& updateId, std::vector<uint64_t>& lostId);

    int Run(const std::vector<std::vector<float> >& dets_high, const std::vector<std::vector<float> >& dets_low, int _no/*观测数量*/, int _ns/*状态数量*/, std::map<uint64_t, int>& update_high_ids, std::map<uint64_t, int>& update_low_ids, std::vector<uint64_t>& lostId);

	std::map<uint64_t, std::shared_ptr<T> >& GetStates();


    void AssociateDetectionsToTrackers(const std::vector<std::vector<float> >& detections, int _no/*观测数量*/, int _ns/*状态数量*/, std::map<uint64_t, std::shared_ptr<T> >& tracks, std::map<uint64_t, int>& matched, std::vector<int>& unmatched_det);

    void AssociateDetectionsToTrackersEx(const std::vector<std::vector<float> >& dets_high, const std::vector<std::vector<float> >& dets_low, int _no/*观测数量*/, int _ns/*状态数量*/, std::map<uint64_t, std::shared_ptr<T> >& tracks, std::map<uint64_t, int>& high_matched, std::map<uint64_t, int>& low_matched, std::vector<int>& unmatched_det);

public:

	std::map<uint64_t, std::shared_ptr<T> > m_tracker;

	uint64_t m_countId = 0;//生成物体id的累加值

    int m_isGPU = 0;//默认不使用

    float m_iou_threshold = 0.01;//iou计算之后匹配的最小值
    int m_kMaxCoastCycles = 2;//predict之后，计算几次认为丢失。
    int m_updateValidCount = 3;//更新多少次认为是一个有效数据。

    std::vector<float> m_values;//调整矩阵参数的配置
};


template<class T>
int BaseTracker<T>::Run(const std::vector<std::vector<float> >& detections, int _no/*观测数量*/, int _ns/*状态数量*/, std::vector<uint64_t>& detectionsId, std::map<uint64_t, int>& updateId, std::vector<uint64_t>& lostId)
{
    /*** Predict internal tracks from previous frame ***/
#ifdef _USING_NSIGHT_
    nvtxRangePush("Run Predict");
#endif
    for (auto& track : m_tracker)
    {
        track.second->Predict();
        //std::vector<float> predict;
        //track.second->GetPredictData(predict);
        //SDK_LOG(SDK_INFO, "predict id = %d, data = [%f,%f,%f,%f,%f,%f,%f]",track.first,predict[0], predict[1], predict[2], predict[3], predict[4], predict[5], predict[6]);
    }
#ifdef _USING_NSIGHT_
    nvtxRangePop();
#endif
    if (detections.empty())
    {
        /*** Delete lose tracked tracks ***/
        for (auto it = m_tracker.begin(); it != m_tracker.end();)
        {
            if (it->second->IsLost())
            {
                lostId.push_back(it->first);
                it = m_tracker.erase(it);
            }
            else
            {
                it++;
            }
        }
        return 0;
    }
#ifdef _USING_NSIGHT_
    nvtxRangePush("Run AssociateDetectionsToTrackers");
#endif
    detectionsId.resize(detections.size());
    // Hash-map between track ID and associated detection bounding box
    std::map<uint64_t, int> matched;
    // vector of unassociated detections
    std::vector<int> unmatched_det;

    // return values - matched, unmatched_det
    AssociateDetectionsToTrackers(detections,_no,_ns, m_tracker, matched, unmatched_det);
#ifdef _USING_NSIGHT_
    nvtxRangePop();
#endif
    /*** Update tracks with associated bbox ***/
#ifdef _USING_NSIGHT_
    nvtxRangePush("Run Update");
#endif
    if (m_isGPU == 0)
    {
        for (const auto& match : matched)
        {
            const auto& id = match.first;
            m_tracker[id]->Update(detections[match.second]);
            detectionsId[match.second] = id;
            updateId[id] = match.second;
        }
    }
    else
    {
        int bs = matched.size();
        if (bs > 0)
        {
            int ns = 0;
            int no = _no;
            std::shared_ptr<float> Z = std::shared_ptr<float>(new float[bs * no], [](float* p) {if (p) delete[] p; p = nullptr; });
            std::shared_ptr<float> X = std::shared_ptr<float>(new float[bs * _ns], [](float* p) {if (p) delete[] p; p = nullptr; });
            std::shared_ptr<float> P = std::shared_ptr<float>(new float[bs * _ns * _ns], [](float* p) {if (p) delete[] p; p = nullptr; });
            std::shared_ptr<float> R = std::shared_ptr<float>(new float[no * no], [](float* p) {if (p) delete[] p; p = nullptr; });
            std::shared_ptr<float> HX = std::shared_ptr<float>(new float[bs * no], [](float* p) {if (p) delete[] p; p = nullptr; });
            int bs_i = 0;
            for (const auto& match : matched)
            {
                const auto& id = match.first;
                std::vector<float> cre_det;
                m_tracker[id]->UpdateDataCheck(detections[match.second], cre_det);
                float* ptr_Z = Z.get() + bs_i * no;
                for (int i = 0; i < no; i++)
                {
                    ptr_Z[i] = cre_det[i];
                }
                memcpy(X.get() + bs_i * _ns, m_tracker[id]->GetStatesXPtr(), _ns * sizeof(float));
                memcpy(P.get() + bs_i * _ns * _ns, m_tracker[id]->GetPredictPtr(), _ns * _ns * sizeof(float));
                //X.get()[bs_i] = m_tracker[id]->GetStatesXPtr();
                //P.get()[bs_i] = m_tracker[id]->GetPredictPtr();
                float* ptr_HX = HX.get() + bs_i * no;
                memcpy(ptr_HX, m_tracker[id]->GetStatesXPtr(), no * sizeof(float));
                if (ns == 0)
                {
                    ns = m_tracker[id]->GetStatesNum();
                    float* _r = m_tracker[id]->GetRPtr();
                    memcpy(R.get(), _r, no * no * sizeof(float));
                }
                bs_i++;
                m_tracker[id]->UpdateHit();//gpu时需要update里面的变量
                detectionsId[match.second] = id;
                updateId[id] = match.second;
            }
            //SDK_LOG(SDK_INFO, "Z = [%s]", GetMatrixStr(Z.get(), no, bs).c_str());
            //SDK_LOG(SDK_INFO, "X = [%s]", GetMatrixStr(X.get(), ns, bs).c_str());
            //SDK_LOG(SDK_INFO, "P = [%s]", GetMatrixStr(P.get(), ns * ns, bs).c_str());
            //SDK_LOG(SDK_INFO, "R = [%s]", GetMatrixStr(R.get(), no, no).c_str());
            //SDK_LOG(SDK_INFO, "HX = [%s]", GetMatrixStr(HX.get(),no,bs).c_str());
#ifdef _KF_IOU_CUDA_
#ifdef _USING_NSIGHT_
            nvtxRangePush("kalman_update_batch");
#endif
            kalman_update_batch(Z.get(), X.get(), P.get(), R.get(), HX.get(), bs, ns, no);
#ifdef _USING_NSIGHT_
            nvtxRangePop();
#endif
#endif
            //SDK_LOG(SDK_INFO, "after X = [%s]", GetMatrixStr(X.get(), ns, bs).c_str());
            //SDK_LOG(SDK_INFO, "after P = [%s]", GetMatrixStr(P.get(), ns * ns, bs).c_str());
            bs_i = 0;
            for (const auto& match : matched)
            {
                const auto& id = match.first;
                memcpy(m_tracker[id]->GetStatesXPtr(), X.get() + bs_i * _ns, _ns * sizeof(float));
                memcpy(m_tracker[id]->GetPredictPtr(), P.get() + bs_i * _ns * _ns, _ns * _ns * sizeof(float));
                bs_i++;
            }
        }
    }
#ifdef _USING_NSIGHT_
    nvtxRangePop();
#endif
    /*** Create new tracks for unmatched detections ***/
    for (const auto& det : unmatched_det)
    {
        std::shared_ptr<T> trackPtr = std::make_shared<T>();
        trackPtr->Init(detections[det]);
        trackPtr->SetIouThreshold(m_iou_threshold);
        trackPtr->SetMaxCoastCycles(m_kMaxCoastCycles);
        trackPtr->SetValidUpdateCount(m_updateValidCount);
        trackPtr->SetValues(m_values);
        // Create new track and generate new ID
        uint64_t newId = ++m_countId;
        m_tracker[newId] = trackPtr;
        detectionsId[det] = newId;
        updateId[newId] = det;
    }

    /*** Delete lose tracked tracks ***/
    for (auto it = m_tracker.begin(); it != m_tracker.end();)
    {
        if (it->second->IsLost())
        {
            lostId.push_back(it->first);
            it = m_tracker.erase(it);
        }
        else
        {
            it++;
        }
    }
    return 0;
}

template<class T>
std::map<uint64_t, std::shared_ptr<T> >& BaseTracker<T>::GetStates()
{
    return m_tracker;
}

template<class T>
void BaseTracker<T>::AssociateDetectionsToTrackers(const std::vector<std::vector<float> >& detections, int _no/*观测数量*/, int _ns/*状态数量*/, std::map<uint64_t, std::shared_ptr<T> >& tracks, std::map<uint64_t, int>& matched, std::vector<int>& unmatched_det)
{
    if (tracks.empty())
    {
        //不做匹配
        for (int i = 0; i < detections.size(); i++)
        {
            unmatched_det.push_back(i);
        }
        return;
    }

    std::vector<std::vector<float>> iou_matrix;
    // resize IOU matrix based on number of detection and tracks
    iou_matrix.resize(detections.size(), std::vector<float>(tracks.size()));

    std::vector<std::vector<float>> association;
    // resize association matrix based on number of detection and tracks
    association.resize(detections.size(), std::vector<float>(tracks.size()));

    if (m_isGPU == 0)
    {
        // row - detection, column - tracks
        for (size_t i = 0; i < detections.size(); i++)
        {
            size_t j = 0;
            for (const auto& trk : tracks)
            {
                iou_matrix[i][j] = trk.second->CalculateIou(detections[i]);
                j++;
            }
        }
        //std::string dete_str = GetMatrixStr(detections, detections.size(), detections.size() > 0 ? detections[0].size() : 0);
        //std::vector<std::vector<float> > tracker_states;
        //for (auto& iter : tracks)
        //{
        //    std::vector<float> measure;
        //    if (iter.second->GetMeasureData(measure) == 0)
        //    {
        //        tracker_states.emplace_back(measure);
        //    }
        //    else
        //    {
        //        SDK_LOG(SDK_INFO, "GetMeasureData failed");
        //    }
        //}
        //std::string track_str = GetMatrixStr(tracker_states, tracker_states.size(), tracker_states.size() > 0 ? tracker_states[0].size() : 0);
        //SDK_LOG(SDK_INFO, "detections = [%s]", dete_str.c_str());
        //SDK_LOG(SDK_INFO, "tracker_states = [%s]", track_str.c_str());
    }
    else
    {
        std::vector<std::vector<float> > tracker_states;
        std::vector<int> tracker_type;
        for (auto& iter : tracks)
        {
            std::vector<float> measure;
            int obj_type = 0;
            if (iter.second->GetIouData(measure,obj_type) == 0)
            {
                tracker_states.emplace_back(measure);
            }
            else
            {
                printf("GetMeasureData failed");
            }
            tracker_type.push_back(obj_type);
        }
        int measure_size = _no;
        int tra_size = tracker_states.size() > 0 ? tracker_states[0].size() : 0;
        if (tra_size == 0 || measure_size != tra_size)
            return;
        int detect_size = detections.size() * measure_size;
        int tracker_size = tracker_states.size() * tra_size;
        int iou_size = detections.size() * tracker_states.size();
        std::shared_ptr<float> detect_ptr = std::shared_ptr<float>(new float[detect_size], [](float* p) {if (p) delete[] p; p = nullptr; });
        std::shared_ptr<float> tracker_ptr = std::shared_ptr<float>(new float[tracker_size], [](float* p) {if (p) delete[] p; p = nullptr; });
        std::shared_ptr<float> iou_ptr = std::shared_ptr<float>(new float[iou_size], [](float* p) {if (p) delete[] p; p = nullptr; });
        std::vector<int> det_type;
        for (int i = 0; i < detections.size(); i++)
        {
            std::vector<float> out;
            int ob_type = 0;
            T::MeasureIouData(detections[i], out, ob_type);
            for (int j = 0; j < out.size(); j++)
            {
                detect_ptr.get()[i * measure_size + j] = out[j];
            }
            det_type.push_back(ob_type);
        }
        for (int i = 0; i < tracker_states.size(); i++)
        {
            for (int j = 0; j < tra_size; j++)
            {
                tracker_ptr.get()[i * tra_size + j] = tracker_states[i][j];
            }
        }
        //std::string dete_str = GetMatrixStr(detect_ptr.get(), measure_size, detections.size());
        //std::string track_str = GetMatrixStr(tracker_states, tracker_states.size(), tra_size);
        //SDK_LOG(SDK_INFO, "detections = [%s]",dete_str.c_str());
        //SDK_LOG(SDK_INFO, "tracker_states = [%s]", track_str.c_str());
#ifdef _KF_IOU_CUDA_
#ifdef _USING_NSIGHT_
        nvtxRangePush("bev_overlap");
#endif
        bev_overlap(detections.size(), detect_ptr.get(), tracker_states.size(), tracker_ptr.get(), iou_ptr.get());
#ifdef _USING_NSIGHT_
        nvtxRangePop();
#endif
#endif
        for(int i = 0; i < detections.size(); i++)
            for (int j = 0; j < tracker_states.size(); j++)
            {
                float* i_ptr = iou_ptr.get();
                if (det_type[i] == tracker_type[j])
                {
                    iou_matrix[i][j] = i_ptr[i * tracker_states.size() + j];
                }
                else
                {
                    iou_matrix[i][j] = 0.0f;
                }
            }
    }

    // Find association
    //std::string str = GetMatrixStr(iou_matrix, detections.size(), tracks.size());
    //SDK_LOG(SDK_INFO, "iou_matrix = [%s]",str.c_str());
    HungarianMatching(iou_matrix, detections.size(), tracks.size(), association);

    for (size_t i = 0; i < detections.size(); i++)
    {
        bool matched_flag = false;
        size_t j = 0;
        for (auto& trk : tracks)
        {
            if (0 == association[i][j])
            {
                // Filter out matched with low IOU
                //SDK_LOG(SDK_INFO, "match info i = %d,j = %d, iou_matrix = %f, m_iou_threshold = %f",i,j, iou_matrix[i][j], trk.second->m_iou_threshold);
                if (iou_matrix[i][j] >= trk.second->m_iou_threshold)
                {
                    matched[trk.first] = i;
                    trk.second->m_prob = iou_matrix[i][j];
                    matched_flag = true;
                }
                // It builds 1 to 1 association, so we can break from here
                break;
            }
            j++;
        }
        // if detection cannot match with any tracks
        if (!matched_flag) {
            unmatched_det.push_back(i);
        }
    }
}

template<class T>
int BaseTracker<T>::Run(const std::vector<std::vector<float> >& dets_high, const std::vector<std::vector<float> >& dets_low, int _no/*观测数量*/, int _ns/*状态数量*/, std::map<uint64_t, int>& update_high_ids, std::map<uint64_t, int>& update_low_ids, std::vector<uint64_t>& lostId)
{
#ifdef _USING_NSIGHT_
    nvtxRangePush("Run Predict");
#endif
    for (auto& track : m_tracker)
    {
        track.second->Predict();
    }
#ifdef _USING_NSIGHT_
    nvtxRangePop();
#endif
    if (dets_high.empty() && dets_low.empty())
    {
        /*** Delete lose tracked tracks ***/
        for (auto it = m_tracker.begin(); it != m_tracker.end();)
        {
            if (it->second->IsLost())
            {
                lostId.push_back(it->first);
                it = m_tracker.erase(it);
            }
            else
            {
                it++;
            }
        }
        return 0;
    }
#ifdef _USING_NSIGHT_
    nvtxRangePush("Run AssociateDetectionsToTrackers");
#endif
    // Hash-map between track ID and associated detection bounding box
    std::map<uint64_t, int> high_matched;
    std::map<uint64_t, int> low_matched;
    // vector of unassociated detections
    std::vector<int> unmatched_det;

    // return values - matched, unmatched_det
    AssociateDetectionsToTrackersEx(dets_high, dets_low, _no, _ns, m_tracker, high_matched, low_matched, unmatched_det);
#ifdef _USING_NSIGHT_
    nvtxRangePop();
#endif
    /*** Update tracks with associated bbox ***/
#ifdef _USING_NSIGHT_
    nvtxRangePush("Run Update");
#endif
    for (const auto& match : high_matched)
    {
        const auto& id = match.first;
        m_tracker[id]->Update(dets_high[match.second]);
        update_high_ids[id] = match.second;
    }
    for (const auto& match : low_matched)
    {
        const auto& id = match.first;
        m_tracker[id]->Update(dets_low[match.second]);
        update_low_ids[id] = match.second;
    }
#ifdef _USING_NSIGHT_
    nvtxRangePop();
#endif
    /*** Create new tracks for unmatched detections ***/
    for (const auto& det : unmatched_det)
    {
        std::shared_ptr<T> trackPtr = std::make_shared<T>();
        trackPtr->Init(dets_high[det]);
        trackPtr->SetIouThreshold(m_iou_threshold);
        trackPtr->SetMaxCoastCycles(m_kMaxCoastCycles);
        trackPtr->SetValidUpdateCount(m_updateValidCount);
        trackPtr->SetValues(m_values);
        // Create new track and generate new ID
        uint64_t newId = ++m_countId;
        m_tracker[newId] = trackPtr;
        update_high_ids[newId] = det;
    }

    /*** Delete lose tracked tracks ***/
    for (auto it = m_tracker.begin(); it != m_tracker.end();)
    {
        if (it->second->IsLost())
        {
            lostId.push_back(it->first);
            it = m_tracker.erase(it);
        }
        else
        {
            it++;
        }
    }
    return 0;
}

template<class T>
void BaseTracker<T>::AssociateDetectionsToTrackersEx(const std::vector<std::vector<float> >& dets_high, const std::vector<std::vector<float> >& dets_low, int _no/*观测数量*/, int _ns/*状态数量*/, std::map<uint64_t, std::shared_ptr<T> >& tracks, std::map<uint64_t, int>& high_matched, std::map<uint64_t, int>& low_matched, std::vector<int>& unmatched_det)
{
    if (tracks.empty())
    {
        //不做匹配
        for (int i = 0; i < dets_high.size(); i++)
        {
            unmatched_det.push_back(i);
        }
        return;
    }
    if (dets_high.size() > 0)     //高分检测不能为空
    {
        std::vector<std::vector<float>> iou_matrix_high;
        // resize IOU matrix based on number of detection and tracks
        iou_matrix_high.resize(dets_high.size(), std::vector<float>(tracks.size()));

        std::vector<std::vector<float>> association_high;
        // resize association matrix based on number of detection and tracks
        association_high.resize(dets_high.size(), std::vector<float>(tracks.size()));

        for (size_t i = 0; i < dets_high.size(); i++)
        {
            size_t j = 0;
            for (const auto& trk : tracks)
            {
                iou_matrix_high[i][j] = trk.second->CalculateIou(dets_high[i]);
                j++;
            }
        }

        // Find association
        //std::string str = GetMatrixStr(iou_matrix, detections.size(), tracks.size());
        //SDK_LOG(SDK_INFO, "iou_matrix = [%s]",str.c_str());
        HungarianMatching(iou_matrix_high, dets_high.size(), tracks.size(), association_high);

        for (size_t i = 0; i < dets_high.size(); i++)
        {
            bool matched_flag = false;
            size_t j = 0;
            for (auto& trk : tracks)
            {
                if (0 == association_high[i][j])
                {
                    // Filter out matched with low IOU
                    //SDK_LOG(SDK_INFO, "match info i = %d,j = %d, iou_matrix = %f, m_iou_threshold = %f",i,j, iou_matrix[i][j], trk.second->m_iou_threshold);
                    if (iou_matrix_high[i][j] >= trk.second->m_iou_threshold)
                    {
                        high_matched[trk.first] = i;
                        trk.second->m_prob = iou_matrix_high[i][j];
                        matched_flag = true;
                    }
                    // It builds 1 to 1 association, so we can break from here
                    break;
                }
                j++;
            }
            // if detection cannot match with any tracks
            if (!matched_flag) {
                unmatched_det.push_back(i);
            }
        }
    }

    if (dets_low.size() > 0 && tracks.size() - high_matched.size() > 0)  //低分检测不能为空且总跟踪目标减高分检测不能为空
    {
        std::vector<std::vector<float>> iou_matrix_low;
        // resize IOU matrix based on number of detection and tracks
        iou_matrix_low.resize(dets_low.size(), std::vector<float>(tracks.size() - high_matched.size()));

        std::vector<std::vector<float>> association_low;
        // resize association matrix based on number of detection and tracks
        association_low.resize(dets_low.size(), std::vector<float>(tracks.size() - high_matched.size()));

        for (size_t i = 0; i < dets_low.size(); i++)
        {
            size_t j = 0;
            for (const auto& trk : tracks)
            {
                if (high_matched.find(trk.first) == high_matched.end())
                {
                    iou_matrix_low[i][j] = trk.second->CalculateIou(dets_low[i]);
                    j++;
                }
            }
        }

        // Find association
        //std::string str = GetMatrixStr(iou_matrix, detections.size(), tracks.size());
        //SDK_LOG(SDK_INFO, "iou_matrix = [%s]",str.c_str());
        HungarianMatching(iou_matrix_low, dets_low.size(), tracks.size() - high_matched.size(), association_low);

        for (size_t i = 0; i < dets_low.size(); i++)
        {
            bool matched_flag = false;
            size_t j = 0;
            for (auto& trk : tracks)
            {
                if (high_matched.find(trk.first) == high_matched.end())
                {
                    if (0 == association_low[i][j])
                    {
                        // Filter out matched with low IOU
                        //SDK_LOG(SDK_INFO, "match info i = %d,j = %d, iou_matrix = %f, m_iou_threshold = %f",i,j, iou_matrix[i][j], trk.second->m_iou_threshold);
                        if (iou_matrix_low[i][j] >= trk.second->m_iou_threshold)
                        {
                            low_matched[trk.first] = i;
                            trk.second->m_prob = iou_matrix_low[i][j];
                            matched_flag = true;
                        }
                        // It builds 1 to 1 association, so we can break from here
                        break;
                    }
                    j++;
                }
            }
        }
    }


}
