#include "Iou.h"
//#include <iostream>
//#include <vector>
//#include <math.h>
//#include <string.h>
//#include <algorithm>
//#include <opencv2/opencv.hpp>
//#include "caffe/nms.h"
//using namespace std;
//using namespace cv;


void HungarianMatching(const std::vector<std::vector<float>>& iou_matrix,
    size_t nrows, size_t ncols,
    std::vector<std::vector<float>>& association)
{
    Matrix<float> matrix(nrows, ncols);
    // Initialize matrix with IOU values
    for (size_t i = 0; i < nrows; i++) {
        for (size_t j = 0; j < ncols; j++) {
            // Multiply by -1 to find max cost
            if (iou_matrix[i][j] != 0) {
                matrix(i, j) = -iou_matrix[i][j];
            }
            else {
                // TODO: figure out why we have to assign value to get correct result
                matrix(i, j) = 1.0f;
            }
        }
    }

    //    // Display begin matrix state.
    //    for (size_t row = 0 ; row < nrows ; row++) {
    //        for (size_t col = 0 ; col < ncols ; col++) {
    //            std::cout.width(10);
    //            std::cout << matrix(row,col) << ",";
    //        }
    //        std::cout << std::endl;
    //    }
    //    std::cout << std::endl;


        // Apply Kuhn-Munkres algorithm to matrix.
    Munkres<float> m;
    m.solve(matrix);

    //    // Display solved matrix.
    //    for (size_t row = 0 ; row < nrows ; row++) {
    //        for (size_t col = 0 ; col < ncols ; col++) {
    //            std::cout.width(2);
    //            std::cout << matrix(row,col) << ",";
    //        }
    //        std::cout << std::endl;
    //    }
    //    std::cout << std::endl;

    for (size_t i = 0; i < nrows; i++) {
        for (size_t j = 0; j < ncols; j++) {
            association[i][j] = matrix(i, j);
        }
    }
}


struct CVPoint2f
{
    CVPoint2f(float _x, float _y):x(_x),y(_y)
    {
    }
    CVPoint2f(const CVPoint2f& obj)
    {
        x = obj.x;
        y = obj.y;
    }
    float x;
    float y;
};


bool bInBox(const std::vector<CVPoint2f> &vpBoxA, const CVPoint2f &p)
{
    std::vector<CVPoint2f> corners = vpBoxA;
    for(int i = 0; i<vpBoxA.size(); i++)   //01230123
        corners.push_back(vpBoxA[i]);

    std::vector< std::vector<double> > linesA;
    for(int i = 0; i<vpBoxA.size(); i++)
    {
        CVPoint2f p1 = corners[i];
        CVPoint2f p2 = corners[i+1];
        CVPoint2f p3 = corners[i+2];
        double a;
        if(p1.x - p2.x == 0)
            a = -(p1.y - p2.y)/0.0001;
        else
            a = -(p1.y - p2.y)/(p1.x - p2.x);

        double b = 1;
        double c = -a * p2.x - p2.y;
        double d = a*p3.x + b*p3.y + c;

        std::vector<double> line{a,b,c,d};
        linesA.push_back(line);
    }

    for(int i=0; i<linesA.size(); i++)
    {
        std::vector<double > l = linesA[i];
        double y = l[0] * p.x + l[1] * p.y +l[2];
        if(y * l[3] < 0)
            return false;
    }
    return true;
}

double InterSection_2D(const std::vector<CVPoint2f> &vpBoxA, const std::vector<CVPoint2f> &vpBoxB)
{
    double min_x, max_x, min_y, max_y;
    min_x = vpBoxA[0].x;
    max_x = vpBoxA[0].x;
    min_y = vpBoxA[0].y;
    max_y = vpBoxA[0].y;

    for(int i=1; i<vpBoxA.size(); i++)
    {
        CVPoint2f p = vpBoxA[i];
        if(p.x > max_x)
            max_x = p.x;
        if(p.x < min_x)
            min_x = p.x;
        if(p.y > max_y)
            max_y = p.y;
        if(p.y < min_y)
            min_y = p.y;
    }
    for(int i=0; i<vpBoxB.size(); i++)
    {
        CVPoint2f p = vpBoxB[i];
        if(p.x > max_x)
            max_x = p.x;
        if(p.x < min_x)
            min_x = p.x;
        if(p.y > max_y)
            max_y = p.y;
        if(p.y < min_y)
            min_y = p.y;
    }

    //将两个BBox的定点坐标最小值设置为0, 以防止有负数的产生
    std::vector<CVPoint2f> vpBoxAA = vpBoxA;
    std::vector<CVPoint2f> vpBoxBB = vpBoxB;

    //if(min_x < 0 && min_y < 0)
    for(int i=0; i<vpBoxA.size(); i++)
    {
        vpBoxAA[i].x = vpBoxAA[i].x - min_x;
        vpBoxAA[i].y = vpBoxAA[i].y - min_y;
        vpBoxBB[i].x = vpBoxBB[i].x - min_x;
        vpBoxBB[i].y = vpBoxBB[i].y - min_y;
    }

    int imax_x = (int)((max_x - min_x) * 10000);
    int imax_y = (int)((max_y - min_y) * 10000);

    double points_inA = 0, points_inB = 0, points_inAB = 0;
    srand((int)time(0));
    for(int i = 0; i<100000; i++)
    {
        int xx = rand()%(imax_x)+1;  //生成[1, imax_x]之间的整数
        int yy = rand()%(imax_y)+1;  //生成[1, imax_y]之间的整数

        CVPoint2f p((float)xx / 10000.0, (float)yy / 10000.0);
        if( bInBox(vpBoxAA, p) )
            ++points_inA;
        if( bInBox(vpBoxBB, p) )
            ++points_inB;
        if( bInBox(vpBoxAA, p) && bInBox(vpBoxBB, p) )
            ++points_inAB;
    }

    double iou = points_inAB / (points_inA + points_inB - points_inAB);
    //cout<<"points_inA : "<<points_inA<<",  points_inB: "<<points_inB<<" ,points_inAB: "<<points_inAB<<endl;
    return iou;
}
//double CuboidIoU(const Eigen::MatrixXd &truth_poses, const Eigen::MatrixXd &landmark_poses)
double CuboidIoU(const std::vector<float> &truth_poses, const std::vector<float> &landmark_poses)
{

    std::vector<CVPoint2f>  vground_points;
    std::vector<CVPoint2f>  vlandmark_points;
    if(1)  //通过坐标旋转求取groundtruth立方体中 2D-Boundbox四个顶点的坐标
    {
        //double cen_x = truth_poses(0,0);
        //double cen_y = truth_poses(0,1);
        //double len = truth_poses(0,6);
        //double wid = truth_poses(0,7);
        //double yaw = truth_poses(0,5);
        double cen_x = truth_poses[0];
        double cen_y = truth_poses[1];
        double len = truth_poses[6];
        double wid = truth_poses[7];
        double yaw = truth_poses[5];

        double x, y, xx, yy;
        x = cen_x - len;
        y = cen_y - wid;
        xx = (x - cen_x)*cos(yaw) - (y - cen_y)*sin(yaw) + cen_x;
        yy = (x - cen_x)*sin(yaw) + (y - cen_y)*cos(yaw) + cen_y;
        CVPoint2f p0(xx, yy);

        x = cen_x - len;
        y = cen_y + wid;
        xx = (x - cen_x)*cos(yaw) - (y - cen_y)*sin(yaw) + cen_x;
        yy = (x - cen_x)*sin(yaw) + (y - cen_y)*cos(yaw) + cen_y;
        CVPoint2f p1(xx, yy);

        x = cen_x + len;
        y = cen_y + wid;
        xx = (x - cen_x)*cos(yaw) - (y - cen_y)*sin(yaw) + cen_x;
        yy = (x - cen_x)*sin(yaw) + (y - cen_y)*cos(yaw) + cen_y;
        CVPoint2f p2(xx, yy);

        x = cen_x + len;
        y = cen_y - wid;
        xx = (x - cen_x)*cos(yaw) - (y - cen_y)*sin(yaw) + cen_x;
        yy = (x - cen_x)*sin(yaw) + (y - cen_y)*cos(yaw) + cen_y;
        CVPoint2f p3(xx, yy);

        vground_points = {p0, p1, p2, p3};
    }

    if(1)//通过坐标旋转求取landmark中 2D-Boundbox四个顶点的坐标
    {
        //double cen_x = landmark_poses(0,0);
        //double cen_y = landmark_poses(0,1);
        //double len = landmark_poses(0,6);
        //double wid = landmark_poses(0,7);
        //double yaw = landmark_poses(0,5);
        double cen_x = landmark_poses[0];
        double cen_y = landmark_poses[1];
        double len = landmark_poses[6];
        double wid = landmark_poses[7];
        double yaw = landmark_poses[5];

        double x, y, xx, yy;
        x = cen_x - len;
        y = cen_y - wid;
        xx = (x - cen_x)*cos(yaw) - (y - cen_y)*sin(yaw) + cen_x;
        yy = (x - cen_x)*sin(yaw) + (y - cen_y)*cos(yaw) + cen_y;
        CVPoint2f p0(xx, yy);

        x = cen_x - len;
        y = cen_y + wid;
        xx = (x - cen_x)*cos(yaw) - (y - cen_y)*sin(yaw) + cen_x;
        yy = (x - cen_x)*sin(yaw) + (y - cen_y)*cos(yaw) + cen_y;
        CVPoint2f p1(xx, yy);

        x = cen_x + len;
        y = cen_y + wid;
        xx = (x - cen_x)*cos(yaw) - (y - cen_y)*sin(yaw) + cen_x;
        yy = (x - cen_x)*sin(yaw) + (y - cen_y)*cos(yaw) + cen_y;
        CVPoint2f p2(xx, yy);

        x = cen_x + len;
        y = cen_y - wid;
        xx = (x - cen_x)*cos(yaw) - (y - cen_y)*sin(yaw) + cen_x;
        yy = (x - cen_x)*sin(yaw) + (y - cen_y)*cos(yaw) + cen_y;
        CVPoint2f p3(xx, yy);

        vlandmark_points = {p0, p1, p2, p3};
    }

    double iou_2d = InterSection_2D(vlandmark_points, vground_points);
    std::cout << " iou_2d = " << iou_2d << std::endl;
    double iou_3d = 0;
    if (iou_2d > 0) {
        //double tru_minz = truth_poses(0, 2) - truth_poses(0, 8);
        //double tru_maxz = truth_poses(0, 2) + truth_poses(0, 8);

        //double land_minz = landmark_poses(0, 2) - landmark_poses(0, 8);
        //double land_maxz = landmark_poses(0, 2) + landmark_poses(0, 8);

        double tru_minz = truth_poses[2] - truth_poses[8];
        double tru_maxz = truth_poses[2] + truth_poses[8];

        double land_minz = landmark_poses[2] - landmark_poses[8];
        double land_maxz = landmark_poses[2] + landmark_poses[8];
        if (land_maxz <= tru_maxz && land_maxz >= tru_minz) {
            double height_iou = (land_maxz - tru_minz) / (tru_maxz - land_minz);
            iou_3d = iou_2d * height_iou;
        } else if (tru_maxz < land_maxz && tru_maxz > land_minz) {
            double height_iou = (tru_maxz - land_minz) / (land_maxz - tru_minz);
            iou_3d = iou_2d * height_iou;
        }
    }

    return iou_3d;
}


//void main(int argc, char **argv)
//{
//    Eigen::MatrixXd truth_poses(1,9);
//    truth_poses<<-0.4875, -0.798913, -1.125, 0, 0, -1.27409, 0.240477, 0.235315, 0.375;
//    Eigen::MatrixXd landmark_poses(1,9);
//    landmark_poses<<-0.506616, -0.796303, -1.20499, 0, 0, -0.345443, 0.22,  0.22, 0.4 ;
//    double iou_3d = CuboidIoU_once(truth_poses, landmark_poses);
//    cout<<"the iou of two cuboid is: "<<iou_3d<<endl;
//    return;
//}





//cv::Point2f center;

////c++ 多边形求交集
////const int maxn = 300;
//const double eps = 1e-8;
//int dcmp(double x)
//{
//    if(x > eps) return 1;
//    return x < -eps ? -1 : 0;
//}
//struct MyPoint
//{
//    double x, y;
//};
//double cross(MyPoint a,MyPoint b,MyPoint c) ///叉积
//{
//    return (a.x-c.x)*(b.y-c.y)-(b.x-c.x)*(a.y-c.y);
//}
//MyPoint intersection(MyPoint a,MyPoint b,MyPoint c,MyPoint d)
//{
//    MyPoint p = a;
//    double t =((a.x-c.x)*(c.y-d.y)-(a.y-c.y)*(c.x-d.x))/((a.x-b.x)*(c.y-d.y)-(a.y-b.y)*(c.x-d.x));
//    p.x +=(b.x-a.x)*t;
//    p.y +=(b.y-a.y)*t;
//    return p;
//}
////计算多边形面积
//double PolygonArea(MyPoint p[], int n)
//{
//    if(n < 3) return 0.0;
//    double s = p[0].y * (p[n - 1].x - p[1].x);
//    p[n] = p[0];
//    for(int i = 1; i < n; ++ i)
//        s += p[i].y * (p[i - 1].x - p[i + 1].x);
//    return fabs(s * 0.5);
//}
//double CPIA(MyPoint a[], MyPoint b[], int na, int nb)//ConvexPolygonIntersectArea
//{
//    MyPoint p[20], tmp[20];
//    int tn, sflag, eflag;
//    a[na] = a[0], b[nb] = b[0];
//    memcpy(p,b,sizeof(MyPoint)*(nb + 1));
//    for(int i = 0; i < na && nb > 2; i++)
//    {
//        sflag = dcmp(cross(a[i + 1], p[0],a[i]));
//        for(int j = tn = 0; j < nb; j++, sflag = eflag)
//        {
//            if(sflag>=0) tmp[tn++] = p[j];
//            eflag = dcmp(cross(a[i + 1], p[j + 1],a[i]));
//            if((sflag ^ eflag) == -2)
//                tmp[tn++] = intersection(a[i], a[i + 1], p[j], p[j + 1]); ///求交点
//        }
//        memcpy(p, tmp, sizeof(MyPoint) * tn);
//        nb = tn, p[nb] = p[0];
//    }
//    if(nb < 3) return 0.0;
//    return PolygonArea(p, nb);
//}
//double SPIA(MyPoint a[], MyPoint b[], int na, int nb)///SimplePolygonIntersectArea 调用此函数
//{
//    int i, j;
//    MyPoint t1[4], t2[4];
//    double res = 0, num1, num2;
//    a[na] = t1[0] = a[0], b[nb] = t2[0] = b[0];
//
//    for(i = 2; i < na; i++)
//    {
//        t1[1] = a[i-1], t1[2] = a[i];
//        num1 = dcmp(cross(t1[1], t1[2],t1[0]));
//        if(num1 < 0) swap(t1[1], t1[2]);
//
//        for(j = 2; j < nb; j++)
//        {
//
//            t2[1] = b[j - 1], t2[2] = b[j];
//            num2 = dcmp(cross(t2[1], t2[2],t2[0]));
//            if(num2 < 0) swap(t2[1], t2[2]);
//            res += CPIA(t1, t2, 3, 3) * num1 * num2;
//        }
//    }
//    return res;
//}
////c++ 多边形求交集
//
//double IOU(const proposal_type & r1, const proposal_type & r2)
//{
//    double inter = intersection_area(r1,r2);
//
//    double o = inter / (calcularea(r1) + calcularea(r2) - inter);
//
//    return (o >= 0) ? o : 0;
//}
//
//double intersection_area(const proposal_type & r1, const proposal_type & r2){
//    MyPoint p1[10],p2[10];
//
//    p1[0].x=r1.x2;
//    p1[0].y=r1.y2;
//    p1[1].x=r1.x3;
//    p1[1].y=r1.y3;
//    p1[2].x=r1.x4;
//    p1[2].y=r1.y4;
//    p1[3].x=r1.x1;
//    p1[3].y=r1.y1;
//
//    p2[0].x=r2.x2;
//    p2[0].y=r2.y2;
//    p2[1].x=r2.x3;
//    p2[1].y=r2.y3;
//    p2[2].x=r2.x4;
//    p2[2].y=r2.y4;
//    p2[3].x=r2.x1;
//    p2[3].y=r2.y1;
//    double area = SPIA(p1, p2, 4, 4);
//    return area;
//}
//
//double calcularea(const proposal_type & r){
//    float d12=sqrt(pow(r.x2-r.x1,2)+pow(r.y2-r.y1,2));
//    float d14=sqrt(pow(r.x4-r.x1,2)+pow(r.y4-r.y1,2));
//    float d24=sqrt(pow(r.x2-r.x4,2)+pow(r.y2-r.y4,2));
//    float d32=sqrt(pow(r.x2-r.x3,2)+pow(r.y2-r.y3,2));
//    float d34=sqrt(pow(r.x3-r.x4,2)+pow(r.y3-r.y4,2));
//    float p1=(d12+d14+d24)/2;
//    float p2=(d24+d32+d34)/2;
//    float s1=sqrt(p1*(p1-d12)*(p1-d14)*(p1-d24));
//    float s2=sqrt(p2*(p2-d32)*(p2-d34)*(p2-d24));
//    return s1+s2;
//}
//
//bool cmpr(const proposal_type &a,const proposal_type &b){
//    return a.score > b.score;
//}
//
//void nms(vector<proposal_type>& proposals, const double nms_threshold)
//{
//
//    //按分数排序
//    sort(proposals.begin(),proposals.end(),cmpr);
//    //标志，false代表留下，true代表扔掉
//    vector<bool> del(proposals.size(), false);
//    for(size_t i = 0; i < proposals.size(); i++){
//        if(!del[i]){
//            // std::cout<<scores[i]<<std::endl;
//            for(size_t j = i+1; j < proposals.size()-1; j++){
//                if(!del[j] && IOU(proposals[i], proposals[j]) > nms_threshold){
//                    del[j] = true;//IOU大于阈值，扔掉
//                }
//            }
//        }
//    }
///*for(int i=0;i<del.size();i++){
//cout<<del[i]<<" ";
//}*/
//vector<proposal_type> new_proposals;
///*for(const auto i : index){
//    if(!del[i]) new_proposals.push_back(proposals[i]);
//}*/
//for (int i = 0; i < proposals.size(); i++) {
//    if (!del[i]) new_proposals.push_back(proposals[i]);
//}
////cout<<new_proposals.size()<<endl;
//proposals.clear();
//vector<proposal_type>().swap(proposals);
//proposals = new_proposals;
////    cout<<proposals.size()<<endl;
//    // scores.clear();
//    //vector<float>().swap(scores);
//}

