# Author: Xinshuo Weng
# email: xinshuo.weng@gmail.com
import math
import numpy as np
from kalmanFilter_gpu_class import KalmanFilter_GPU
#from filterpy.kalman import KalmanFilter
from sklearn.decomposition import PCA
import time

NUM_FRAME = 5
DIST_THRED = 0.5

def center_rot_y_f(points):

    x_list = points[:, 0]
    y_list = points[:, 1]

    pca = PCA(n_components=2)
    pca.fit(points)
    PCA(copy=True, n_components=2, whiten=False)

    orientation = pca.components_[0]

    dx = x_list[-1] - x_list[0]
    dy = y_list[-1] - y_list[0]
    alpha = np.dot(np.array([dx, dy]), orientation)

    if alpha < 0:
        orientation = -orientation

    center_rot_y = math.atan2(orientation[1], orientation[0])

    return center_rot_y


class KalmanBoxTracker(object):
    """
    This class represents the internel state of individual tracked objects observed as bbox.
    """
    count = 0

    def __init__(self, bbox3D, info):
        """
        Initialises a tracker using initial bounding box.
        """
        # define constant velocity model
        self.kf = KalmanFilter_GPU(dim_x=10, dim_z=7)
        self.kf.F = np.array([[1, 0, 0, 0, 0, 0, 0, 1, 0, 0],  # state transition matrix
                              [0, 1, 0, 0, 0, 0, 0, 0, 1, 0],
                              [0, 0, 1, 0, 0, 0, 0, 0, 0, 1],
                              [0, 0, 0, 1, 0, 0, 0, 0, 0, 0],
                              [0, 0, 0, 0, 1, 0, 0, 0, 0, 0],
                              [0, 0, 0, 0, 0, 1, 0, 0, 0, 0],
                              [0, 0, 0, 0, 0, 0, 1, 0, 0, 0],
                              [0, 0, 0, 0, 0, 0, 0, 1, 0, 0],
                              [0, 0, 0, 0, 0, 0, 0, 0, 1, 0],
                              [0, 0, 0, 0, 0, 0, 0, 0, 0, 1]])

        self.kf.H = np.array([[1, 0, 0, 0, 0, 0, 0, 0, 0, 0],  # measurement function,
                              [0, 1, 0, 0, 0, 0, 0, 0, 0, 0],
                              [0, 0, 1, 0, 0, 0, 0, 0, 0, 0],
                              [0, 0, 0, 1, 0, 0, 0, 0, 0, 0],
                              [0, 0, 0, 0, 1, 0, 0, 0, 0, 0],
                              [0, 0, 0, 0, 0, 1, 0, 0, 0, 0],
                              [0, 0, 0, 0, 0, 0, 1, 0, 0, 0]])

        self.kf.F_blindspot = np.array([[1, 0, 0, 0, 0, 0, 0, 0, 0, 0],  # state transition matrix
                                        [0, 1, 0, 0, 0, 0, 0, 0, 0.3, 0],
                                        [0, 0, 1, 0, 0, 0, 0, 0, 0, 0.],
                                        [0, 0, 0, 1, 0, 0, 0, 0, 0, 0],
                                        [0, 0, 0, 0, 1, 0, 0, 0, 0, 0],
                                        [0, 0, 0, 0, 0, 1, 0, 0, 0, 0],
                                        [0, 0, 0, 0, 0, 0, 1, 0, 0, 0],
                                        [0, 0, 0, 0, 0, 0, 0, 1, 0, 0],
                                        [0, 0, 0, 0, 0, 0, 0, 0, 1, 0],
                                        [0, 0, 0, 0, 0, 0, 0, 0, 0, 1]])

        self.kf.P[:, :] *= 1000.
        self.kf.P *= 10.

        self.kf.Q[:, :] *= 0.01
        self.kf.x[:7] = bbox3D.reshape((7, 1))

        self.time_since_update = 0
        self.id = KalmanBoxTracker.count
        self.create_stamp = info[0]
        KalmanBoxTracker.count += 1
        self.history = []
        self.hits = 1  # number of total hits including the first detection
        # one update, one hit

#        self.hit_streak = 1  # number of continuing hit considering the first detection
        self.first_continuing_hit = 1
        self.still_first = True
        self.age = 0
        self.info = info  # other info associated
        self.type_history = [info[1]]
        self.color_history = [info[-1]]
        self.license_plate_number_history = [info[-2]]
        self.name_history = [info[-3]]
        last_x_y = [self.kf.x[0][0], self.kf.x[1][0]]
        self.last_x_y_list = [last_x_y]
        self.blind_stay = 0


    def correct_angle(self, bbox3D):
        """
        correct the angles of self.kf.x and bbox3d

        """

        t = time.time()

        points = self.last_x_y_list[:]
        points = np.array(points)
        #points = np.squeeze(points, axis=(2,))
        x_list = points[:, 0]
        y_list = points[:, 1]
        if len(points) >= NUM_FRAME and abs(x_list[-1] - x_list[0]) > DIST_THRED or abs(y_list[-1] - y_list[0]) > DIST_THRED:
            center_rot_y = center_rot_y_f(points)

            rot_y = bbox3D[3]
            if rot_y > np.pi: rot_y -= int((rot_y + np.pi)/(2*np.pi)) * np.pi * 2
            if rot_y < -np.pi: rot_y += int((np.pi - rot_y)/(2*np.pi)) * np.pi * 2

            if abs(center_rot_y - rot_y) > np.pi / 2.0 and abs(
                    center_rot_y - rot_y) < np.pi * 3 / 2.0:  # if the angle of two theta is not acute angle
                rot_y += np.pi
                if rot_y > np.pi: rot_y -= np.pi * 2  # make the theta still in the range
                if rot_y < -np.pi: rot_y += np.pi * 2

            #bbox3D[3] = rot_y
            bbox3D[3] = center_rot_y

        t3_0 = time.time() - t

        t = time.time()


        self.time_since_update = 0
        self.history = []
        self.hits += 1
        if self.still_first:
            self.first_continuing_hit += 1  # number of continuing hit in the fist time

        ######################### orientation correction
        if self.kf.x[3] >= np.pi: self.kf.x[3] -= np.pi * 2  # make the theta still in the range
        if self.kf.x[3] < -np.pi: self.kf.x[3] += np.pi * 2

        new_theta = bbox3D[3]
        if new_theta >= np.pi: new_theta -= np.pi * 2  # make the theta still in the range
        if new_theta < -np.pi: new_theta += np.pi * 2
        bbox3D[3] = new_theta

        predicted_theta = self.kf.x[3]
        if abs(new_theta - predicted_theta) > np.pi / 2.0 and abs(
                new_theta - predicted_theta) < np.pi * 3 / 2.0:  # if the angle of two theta is not acute angle
            self.kf.x[3] += np.pi
            if self.kf.x[3] > np.pi: self.kf.x[3] -= np.pi * 2  # make the theta still in the range
            if self.kf.x[3] < -np.pi: self.kf.x[3] += np.pi * 2

        # now the angle is acute: < 90 or > 270, convert the case of > 270 to < 90
        if abs(new_theta - self.kf.x[3]) >= np.pi * 3 / 2.0:
            if new_theta > 0:
                self.kf.x[3] += np.pi * 2
            else:
                self.kf.x[3] -= np.pi * 2

        t3_1 = time.time() - t


        return bbox3D, t3_0, t3_1


    def assign_info(self, info):

        if self.kf.x[3] >= np.pi: self.kf.x[3] -= np.pi * 2  # make the theta still in the rage
        if self.kf.x[3] < -np.pi: self.kf.x[3] += np.pi * 2
        self.info = info

    def update_single(self, bbox3D, info):
        """
        Updates the state vector with observed bbox.
        """
        # 1. correct_angle
        bbox3D, t3_0, t3_1 = self.correct_angle(bbox3D)
        # TIMING PART 2
        s = time.time()

        self.kf.update(bbox3D)
        t3_2 = time.time() - s
        # 3. assign_info
        self.assign_info(info)
        return t3_0, t3_1, t3_2

    def predict(self):
        """
        Advances the state vector and returns the predicted bounding box estimate.
        """
        if self.time_since_update == 0:
            last_x_y = [self.kf.x[0][0], self.kf.x[1][0]]
            self.last_x_y_list.append(last_x_y)
            if len(self.last_x_y_list) > NUM_FRAME:
                del(self.last_x_y_list[0])

        self.kf.predict()
        if self.kf.x[3] >= np.pi: self.kf.x[3] -= np.pi * 2
        if self.kf.x[3] < -np.pi: self.kf.x[3] += np.pi * 2

        self.age += 1
        if (self.time_since_update > 0):
#            self.hit_streak = 0
            self.still_first = False
        self.time_since_update += 1
        self.history.append(self.kf.x)
        return self.kf.x

    def get_state(self):
        """
        Returns the current bounding box estimate.
        """
        return self.kf.x[:7].reshape((7,))

    def get_orientation(self):
        return self.kf.x[7:].reshape((3,))




