#!/usr/bin/env python
# encoding: utf-8
import os
import threading
import rospy
import time
import math
import yaml
from concurrent.futures import ThreadPoolExecutor, thread
from Queue import Queue
from easydict import EasyDict
import numpy as np
from jfx_common_msgs.msg import det_tracking, det_tracking_array, InferReses
from jsk_recognition_msgs.msg import BoundingBoxArray, BoundingBox
from visualization_msgs.msg import MarkerArray, Marker
from AB3DMOT_libs.model import AB3DMOT
from pyquaternion import Quaternion

from utils.getloc_coll import get_loc, get_world_loc, get_camera_world_loc, Inverse_v2
from utils.compute_yaw import compute_yaw

import copy
import logging
logging.basicConfig()


def color_name2RGB(color_name):
    if color_name == 'red':
        return 1.0,0,0
    elif color_name == 'orange':
        return 1.0,165.0/255,0
    elif color_name == 'yellow':
        return 1.0,1.0,0
    elif color_name == 'green':
        return 0,1.0,0
    elif color_name == 'cyan':
        return 0,1.0,1.0
    elif color_name == 'blue':
        return 0,0,1.0
    elif color_name == 'violet':
        return 238.0/255,130.0/255,238.0/255
    elif color_name == 'black':
        return 0,0,0
    elif color_name == 'grey':
        return 190.0/255,190.0/255,190.0/255
    elif color_name == 'white':
        return 1.0,1.0,1.0
    else:
        return 1.0,1.0,1.0

def to360(rot_y):
    if rot_y > 0:
        rot_y = 2 * np.pi - rot_y
    else:
        rot_y = -rot_y
    return rot_y


def rotz(t):
  """Rotation about the z-axis."""
  c = np.cos(t)
  s = np.sin(t)
  return np.array([[c, -s,  0],
                    [s,  c,  0],
                    [0,  0,  1]])

def init_pubs():
  global cloud_pub, tracking_res_pub, bbox_pub, marker_bbox_pub, marker_pub
  tracking_res_pub = rospy.Publisher("/TrackingRes", det_tracking_array, queue_size=10)
#   bbox_pub = rospy.Publisher("/TrackingBbox1", BoundingBoxArray, queue_size=10)
#   marker_bbox_pub = rospy.Publisher("/MarkerTrackingBbox", MarkerArray, queue_size=10)
#   marker_pub = rospy.Publisher("/TrackingLoc", MarkerArray, queue_size=10)


class Tracker:
    def __init__(self):
        self.queue = Queue(maxsize=0)
        self.index = -1
        self.thread = None

def camera_det_to_lidar_det(inferReses):
    # print("into camera_det_to_lidar_det")

    if len(inferReses.reses) == 0:
        print("null reses")
        return

    # ---- chenkun
    # range num could get from config file, now set as constant 4.
    # for i in range(4):
        # construct four reses_queue for different tracker, using producer & consumer model
    for camera_det_msg in inferReses.reses:
        if not reses_queue_list.get(camera_det_msg.index):
            tracker = Tracker()
            tracker.index = camera_det_msg.index
            tracker.thread = threading.Thread(target=single_reses_process, args=(camera_det_msg.index,))
            tracker.thread.start()
            reses_queue_list[camera_det_msg.index] = tracker
        # print("-----")
        # print(camera_det_msg)
        # print("-----")
        reses_queue_list[camera_det_msg.index].queue.put((inferReses.header, camera_det_msg))
    
    # ---- chenkun


def single_reses_process(device):
    while (True):
        # queue : {index, results}
        if not reses_queue_list.get(device):
            print("1.device:%d" %device)
            time.sleep(0.01)
            continue
        if reses_queue_list[device].queue.empty() == True:
            print("2.device:%d" %device)
            time.sleep(0.01)
            continue
        resesPair = reses_queue_list[device].queue.get()
        reses = resesPair[1]
        if len(reses.results) == 0:
            print("3.device:%d" %device)
            time.sleep(0.01)
            continue
        # timestamp = int(header.stamp.to_sec() * 1000)
        timestamp = int(resesPair[0].stamp.to_nsec() / 1000000)
        # print("---single_reses_process,thread=%d, device=%d, ts=%d, queue size=%d"%(threading.currentThread().ident, device, timestamp, reses_queue_list[device].queue.qsize()))
        # print({timestamp, obj})
        # print("===results: ", results)

        tstart = time.time() * 1000

        msgs = det_tracking_array()
        for result in reses.results:
            prob = result.prob
            num_class = result.num_class
            msg = det_tracking()
            msg.frame = int(timestamp)
            msg.x = float(result.bottom_center_point[0])
            msg.y = float(result.bottom_center_point[1])
            msg.z = 0. #-7.
            msg.score = float(prob)
            msg.color_name = str(device)

            if num_class == 1 or num_class == 13 or num_class == 16 or num_class == 18:
                msg.h = 1.665
                msg.w = 1.865
                msg.l = 4.346
                msg.rot_y = 0.
                msg.type = 1
                msg.name = 'Car'
            elif num_class == 7 or num_class == 11 or num_class == 12 or num_class == 15 or num_class == 17:
                msg.h = 2.7
                msg.w = 2.5
                msg.l = 9.0
                msg.rot_y = 0.
                msg.type = 2
                msg.name = 'Truck'
            elif num_class == 8 or num_class == 9 or num_class == 10 or num_class == 19 or num_class == 20:
                msg.h = 3.4
                msg.w = 2.76
                msg.l = 15
                msg.rot_y = 0.
                msg.type = 3
                msg.name = 'Bus'
            elif num_class == 0:
                msg.h = 1.64
                msg.w = 0.65
                msg.l = 0.565
                msg.rot_y = 0.
                msg.type = 4
                msg.name = 'Person'
            elif num_class == 2:
                msg.h = 1.4
                msg.w = 0.72
                msg.l = 1.71
                msg.rot_y = 0.
                msg.type = 5
                msg.name = 'Bicycle'
            elif num_class == 3:
                msg.h = 1.4
                msg.w = 0.72
                msg.l = 1.71
                msg.rot_y = 0.
                msg.type = 5
                msg.name = 'Motorbike'
            else:
                continue

            msgs.array.append(msg)
        send_msgs(device, msgs)

        dt = time.time() * 1000 - tstart
        print("dt: ", dt, device, len(reses.results), reses_queue_list[device].queue.qsize())


def send_msgs(device, msgs):
    if len(msgs.array) == 0:
        #cloud_pub.publish(msgs.cloud)
        # bbox_arr = BoundingBoxArray()
        # bbox_arr.header.frame_id = "Pandar64"
        # bbox_pub.publish(bbox_arr)

        # marker_bbox_arr = MarkerArray()
        # marker_bbox_pub.publish(marker_bbox_arr)

        # marker_arr = MarkerArray()
        # marker_pub.publish(marker_arr)
        return

    msgs_to_send = det_tracking_array()
    
    # t0 = time.time()
    timestamp = msgs.array[0].frame

    # t0_0 = time.time()
    multi_trackers = mot_tracker[device].predict(timestamp)
    # s1_0 = time.time() - t0_0

    # s1_1 = 0
    # if len(multi_trackers) > 0:
    #     for trackers in multi_trackers:
    #         for tracker in trackers:
    #             msg = det_tracking()
    #             msg.h = float(tracker[0])
    #             msg.w = float(tracker[1])
    #             msg.l = float(tracker[2])
    #             msg.x = float(tracker[3])
    #             msg.y = float(tracker[4])
    #             msg.z = float(tracker[5])
    #             msg.rot_y = float(tracker[6])
    #             msg.obj_id = int(tracker[7])
    #             msg.v_x = float(tracker[8])
    #             msg.v_y = float(tracker[9])
    #             msg.v_z = float(tracker[10])
    #             msg.frame = int(tracker[11])
    #             msg.type = int(tracker[12])
    #             msg.score = float(tracker[13])
    #             msg.name = str(tracker[14])
    #             msg.license_plate_number = str(tracker[15])
    #             msg.color_name = str(tracker[16])

    #             t0_1 = time.time()
    #             # lidar_loc, out_BL, out_center_BL = get_loc([msg.x, msg.y, msg.z, msg.l, msg.w, msg.h, msg.rot_y], msg.type, Trans, kitti2origin)
    #             s1_1 += time.time() - t0_1 

    #             # msg.loc_x = lidar_loc[0]
    #             # msg.loc_y = lidar_loc[1]
    #             # msg.loc_z = lidar_loc[2]
    #             # msg.Lat = out_center_BL[0]
    #             # msg.Long = out_center_BL[1]

    #             while msg.rot_y >= np.pi: msg.rot_y -= np.pi * 2
    #             while msg.rot_y < -np.pi: msg.rot_y += np.pi * 2
    #             tmp_rot_y = msg.rot_y
    #             tmp3_rot_y = compute_yaw(tmp_rot_y)
    #             msg.rot_y = tmp3_rot_y

    #             dcpy_msg = copy.deepcopy(msg)
    #             msgs_to_send.array.append(dcpy_msg)
                
    #             save interpolated frames
    #             if save_log:
    #                 log_file.write('%u,%u,%f,%f,%f,%f,%f,%f,%f,%f,%d,%f,%f,%f,%f,%f,%f,%.8f,%.8f,%.8f,%.8f,%s,%s,%s\n' % (
    #                     msg.frame, msg.type, msg.score, msg.h, msg.w, msg.l, msg.x, msg.y, msg.z, msg.rot_y, 
    #                     msg.obj_id, msg.v_x, msg.v_y, msg.v_z,msg.loc_x, msg.loc_y, msg.loc_z, out_BL[0], out_BL[1], out_center_BL[0], out_center_BL[1], msg.name, msg.license_plate_number, msg.color_name))
    # t1 = time.time()
    # s1 = t1 - t0
        
    dets = []
    infos = []

    
    for msg in msgs.array:
        det = [msg.h, msg.w, msg.l, msg.x, msg.y, msg.z, msg.rot_y]
        info = [msg.frame, msg.type, msg.score, msg.name, msg.license_plate_number, msg.color_name]
        dets.append(det)
        infos.append(info)

    dets_all = {'dets': np.asarray(dets), 'info': np.asarray(infos)}

    # t0 = time.time()

    trackers = mot_tracker[device].update(dets_all)

    # t1 = time.time()
    # s2 = t1 - t0

    # t0 = time.time()
    
    # bbox_arr = BoundingBoxArray()
    # bbox_arr.header.frame_id = "Pandar64"
    
    # marker_bbox_arr = MarkerArray()
    # bbox_i = 1

    # marker_arr = MarkerArray()
    # i = 1

    # s3_0 = 0
    # s3_0_gl = 0
    # s3_1 = 0
    for tracker in trackers:
        # t9 = time.time()

        msg = det_tracking()
        msg.h = float(tracker[0])
        msg.w = float(tracker[1])
        msg.l = float(tracker[2])
        msg.x = float(tracker[3])
        msg.y = float(tracker[4])
        msg.z = float(tracker[5])
        msg.rot_y = float(tracker[6])
        msg.obj_id = int(tracker[7])
        msg.v_x = float(tracker[8])
        msg.v_y = float(tracker[9])
        msg.v_z = float(tracker[10])
        msg.frame = int(tracker[11])
        msg.type = int(tracker[12])
        msg.score = float(tracker[13])
        msg.name = str(tracker[14])
        msg.license_plate_number = str(tracker[15])
        msg.color_name = str(tracker[16])

        tt_gl = time.time()
        # 高斯->经纬度 TODO
        tmp_x = camera_gauss[device][0] + msg.x
        tmp_y = camera_gauss[device][1] + msg.y
        out_bottom_center_BL = Inverse_v2([tmp_x, tmp_y])
        msg.Lat = out_bottom_center_BL[0]
        msg.Long = out_bottom_center_BL[1]
        # s3_0_gl += time.time() - tt_gl

        while msg.rot_y >= np.pi: msg.rot_y -= np.pi * 2
        while msg.rot_y < -np.pi: msg.rot_y += np.pi * 2
        tmp_rot_y = msg.rot_y
        tmp3_rot_y = compute_yaw(tmp_rot_y)
        msg.rot_y = tmp3_rot_y

        dcpy_msg = copy.deepcopy(msg)
        msgs_to_send.array.append(dcpy_msg)
        
        msg.rot_y = tmp_rot_y
        # s3_0 += time.time() - t9
       
        # t10 = time.time()

        # bbox = BoundingBox()
        # bbox.label = msg.obj_id
        # bbox.value = 2
        # bbox.header.frame_id = "Pandar64"
        # bbox.dimensions.x = msg.l
        # bbox.dimensions.y = msg.w
        # bbox.dimensions.z = msg.h
        # bbox.pose.position.x = msg.x
        # bbox.pose.position.y = msg.y
        # bbox.pose.position.z = msg.z
        # R = rotz(tmp_rot_y)
        # R_quat= list(Quaternion(matrix=R))
        # bbox.pose.orientation.x = R_quat[1]
        # bbox.pose.orientation.y = R_quat[2]
        # bbox.pose.orientation.z = R_quat[3]
        # bbox.pose.orientation.w = R_quat[0]
        # bbox_arr.boxes.append(bbox)

        # marker_bbox = Marker()
        # marker_bbox.id = bbox_i
        # marker_bbox.header.frame_id = "Pandar64"
        # marker_bbox.action = Marker.ADD
        # marker_bbox.type = Marker.CUBE
        # marker_bbox.pose.position.x = msg.x
        # marker_bbox.pose.position.y = msg.y
        # marker_bbox.pose.position.z = msg.z
        # R = rotz(tmp_rot_y)
        # R_quat= list(Quaternion(matrix=R))
        # marker_bbox.pose.orientation.x = R_quat[1]
        # marker_bbox.pose.orientation.y = R_quat[2]
        # marker_bbox.pose.orientation.z = R_quat[3]
        # marker_bbox.pose.orientation.w = R_quat[0]
        # marker_bbox.scale.x = msg.l
        # marker_bbox.scale.y = msg.w
        # marker_bbox.scale.z = msg.h
        # RGB = color_name2RGB(msg.color_name)
        # marker_bbox.color.a = 0.8
        # marker_bbox.color.r = RGB[0]
        # marker_bbox.color.g = RGB[1]
        # marker_bbox.color.b = RGB[2]
        # marker_bbox.lifetime = rospy.Duration(0.5)
        # marker_bbox_arr.markers.append(marker_bbox)
        # bbox_i += 1

        # if msg.type != 4:
        #     marker = Marker()
        #     marker.id = i
        #     marker.header.frame_id = "Pandar64"
        #     marker.action = Marker.ADD
        #     marker.type = Marker.SPHERE
        #     marker.pose.position.x = msg.x
        #     marker.pose.position.y = msg.y
        #     marker.pose.position.z = msg.z + msg.h / 2
        #     marker.pose.orientation.x = 0.0
        #     marker.pose.orientation.y = 0.0
        #     marker.pose.orientation.z = 0.0
        #     marker.pose.orientation.w = 1.0
        #     marker.scale.x = 0.5
        #     marker.scale.y = 0.5
        #     marker.scale.z = 0.5
        #     marker.color.a = 1.0
        #     marker.color.r = 1.0
        #     marker.color.g = 1.0
        #     marker.color.b = 1.0
        #     marker.lifetime = rospy.Duration(0.5)
        #     marker_arr.markers.append(marker)
        #     i += 1

        #     t_marker = Marker()
        #     t_marker.id = i
        #     t_marker.header.frame_id = "Pandar64"
        #     t_marker.action = Marker.ADD
        #     t_marker.type = Marker.TEXT_VIEW_FACING
        #     t_marker.pose.position.x = msg.x
        #     t_marker.pose.position.y = msg.y
        #     t_marker.pose.position.z = msg.z + msg.h / 2 + 0.3
        #     t_marker.pose.orientation.x = 0.0
        #     t_marker.pose.orientation.y = 0.0
        #     t_marker.pose.orientation.z = 0.0
        #     t_marker.pose.orientation.w = 1.0
        #     t_marker.scale.x = 1.2
        #     t_marker.scale.y = 1.2
        #     t_marker.scale.z = 1.2
        #     t_marker.color.a = 1.0
        #     t_marker.color.r = 0
        #     t_marker.color.g = 1.0
        #     t_marker.color.b = 0
        #     t_marker.lifetime = rospy.Duration(0.5)
        #     t_marker.text = 'id:{0},type:{1},type_name:{2},v:{3:.2f}m/s,No:{4}\ncolor_name:{5},timestamp:{6}'.format(msg.obj_id, msg.type, msg.name, math.sqrt(msg.v_x*msg.v_x+msg.v_y*msg.v_y)*10, msg.license_plate_number, msg.color_name, msg.frame)
        #     marker_arr.markers.append(t_marker)
        #     i += 1
        
    if len(trackers) > 0:
        tmp = int(trackers[0][11])
    else:
        tmp = 0


    # print("publish msg:")
    # print(msgs_to_send)
    tracking_res_pub.publish(msgs_to_send)
    # bbox_pub.publish(bbox_arr)
    # marker_bbox_pub.publish(marker_bbox_arr)
    # marker_pub.publish(marker_arr)

def listener():
    rospy.init_node('lidar_tracking_image_listener', anonymous=False)
    init_pubs()
    rospy.Subscriber("InferReses", InferReses, camera_det_to_lidar_det)
    rospy.spin()


def merge_new_config(config, new_config):
    for key, val in new_config.items():
        if not isinstance(val, dict):
            config[key] = val
            continue
        if key not in config:
            config[key] = EasyDict()
        merge_new_config(config[key], val)
    return config
    

def read_yaml(path):
    with open(path,  'r') as f:
        new_config = yaml.load(f, Loader=yaml.FullLoader)
    config = EasyDict()
    merge_new_config(config=config, new_config=new_config)
    return config


if __name__ == '__main__':
    global reses_queue_list, device_list, mot_tracker, camera_gauss

    # reses_queue_list: {index, track{queue, index, thread}}
    reses_queue_list = {}
    device_list = []
    mot_tracker = []
    camera_gauss = []
    cfg_path = '/data/catkin_ws_xishan/src/rosrv2x/jfx_rvconfig/param/vision_config.yaml'
    print(cfg_path)
    cfg = read_yaml(cfg_path)
    cam1_gauss = []
    cam1_gauss.append(cfg.CameraParams[0].coordinate.vx)
    cam1_gauss.append(cfg.CameraParams[0].coordinate.vy)
    camera_gauss.append(cam1_gauss)
    print("camera_gauss[0]", camera_gauss[0])

    cam2_gauss = []
    cam2_gauss.append(cfg.CameraParams[1].coordinate.vx)
    cam2_gauss.append(cfg.CameraParams[1].coordinate.vy)
    camera_gauss.append(cam2_gauss)
    print("camera_gauss[1]", camera_gauss[1])

    cam3_gauss = []
    cam3_gauss.append(cfg.CameraParams[2].coordinate.vx)
    cam3_gauss.append(cfg.CameraParams[2].coordinate.vy)
    camera_gauss.append(cam3_gauss)
    print("camera_gauss[2]", camera_gauss[2])

    cam4_gauss = []
    cam4_gauss.append(cfg.CameraParams[3].coordinate.vx)
    cam4_gauss.append(cfg.CameraParams[3].coordinate.vy)
    camera_gauss.append(cam4_gauss)
    print("camera_gauss[3]", camera_gauss[3])

    # -----------------
    # according to config file
    for i in range(4):
        device_list.append(i)
        mot_tracker.append(AB3DMOT(max_age=3, min_hits=3, blind_in=[], blind_out=[]))
        print("mot_tracker", mot_tracker[i])

    listener()
