import numpy as np
from gaussian import Convert_v2, Inverse_v2
from numba import jit



def my_compute_box_3d(center, heading, size):
  # P' = T * R * S * p
  x_corners = [-1,1,1,-1,-1,1,1,-1]
  y_corners = [1,1,-1,-1,1,1,-1,-1]
  z_corners = [1,1,1,1,-1,-1,-1,-1]

  tmp = np.vstack([x_corners, y_corners, z_corners]) / 2.0
  corners_3d = np.ones([4,8])
  corners_3d[:3,:] = tmp

  S = np.diag([size[0],size[1],size[2],1])
  rot = rotz(heading)
  Trans = np.zeros([4,4])
  Trans[:3,:3] = rot
  Trans[0,3] = center[0]
  Trans[1,3] = center[1]
  Trans[2,3] = center[2]
  #Trans[2,3] = center[2] - size[2]/2

  tmp4x4 = np.dot(S,corners_3d)
  world_corners_3d = np.dot(Trans,tmp4x4)
  
  """
  PREVIOUS:
  kitti2origin = [[-0.001081   , 0.99956106, -0.02960616 ,-0.02960616],
                  [-0.99733779 , 0.001081   , 0.07291202  ,0.07291202],
                  [ 0.07291202 , 0.02960616 , 0.99689885 , 0.99689885],
                  [ 0.         , 0.         , 0.         , 1.        ]]

  """                  
  kitti2origin = [[-0.0821106 ,  0.99618338 ,-0.02960616, -0.02960616],
                 [-0.9941427  ,-0.07977548,  0.07291202 , 0.07291202],
                 [ 0.0702719  , 0.0354196 ,  0.99689885,  0.99689885],
                 [ 0        ,  0  ,        0  ,        1        ]]
      
        
  kitti2origin = np.array(kitti2origin)

  world_corners_3d = np.dot(kitti2origin,world_corners_3d)

  return np.transpose(world_corners_3d[:3,:])


@jit
def my_compute_box_3d_jit(corners_3d, center, heading, size):
  # P' = T * R * S * p

  '''
  x_corners = np.array([-1,1,1,-1,-1,1,1,-1])
  y_corners = np.array([1,1,-1,-1,1,1,-1,-1])
  z_corners = np.array([1,1,1,1,-1,-1,-1,-1])

  tmp = np.vstack((x_corners, y_corners, z_corners)) / 2.0
  corners_3d = np.ones([4,8])
  corners_3d[:3,:] = tmp
  '''

  S = np.diag([size[0],size[1],size[2],1])
  rot = rotz(heading)
  Trans = np.zeros([4,4])
  Trans[:3,:3] = rot
  Trans[0,3] = center[0]
  Trans[1,3] = center[1]
  Trans[2,3] = center[2]

  tmp4x4 = np.dot(S,corners_3d)
  world_corners_3d = np.dot(Trans,tmp4x4)

  kitti2origin = np.array([[-0.0821106 ,  0.99618338 ,-0.02960616, -0.02960616],
                 [-0.9941427  ,-0.07977548,  0.07291202 , 0.07291202],
                 [ 0.0702719  , 0.0354196 ,  0.99689885,  0.99689885],
                 [ 0        ,  0  ,        0  ,        1        ]])


  world_corners_3d = np.dot(kitti2origin,world_corners_3d)

  return np.transpose(world_corners_3d[:3,:])


@jit
def rotz(t):
  """Rotation about the z-axis."""
  c = np.cos(t)
  s = np.sin(t)
  return np.array([[c, -s,  0],
                    [s,  c,  0],
                    [0,  0,  1]])


def get_T():
  h = 19.34
  rot = [[-0.330806583166, -0.941249847412, 0.067936882377],
         [0.943696200848, -0.329787790775, 0.026026830077],
         [-0.002092991956, 0.072721630335 ,0.997349977493]]
  rot = np.array(rot)
  BL = [31.2840452,121.1597165]
  XY = Convert_v2(BL) 
  # print(XY)
  Trans = np.zeros([4,4])
  Trans[:3,:3] = rot
  Trans[0,3] = XY[0]
  Trans[1,3] = XY[1]
  Trans[2,3] = h
  Trans[3,3] = 1

  return Trans


def get_loc(boxes_3d, car_type):
  Trans = np.array([[-3.30806583e-01, -9.41249847e-01 , 6.79368824e-02 , 4.06104317e+07],
                     [ 9.43696201e-01, -3.29787791e-01 , 2.60268301e-02 , 3.46304736e+06],
                     [-2.09299196e-03,  7.27216303e-02 , 9.97349977e-01 , 1.93400000e+01],
                     [ 0.00000000e+00,  0.00000000e+00 , 0.00000000e+00 , 1.00000000e+00]])
    


  corners_3d_in = np.array(
          [[ -0.5 , 0.5 , 0.5 , -0.5 , -0.5 , 0.5 , 0.5 , -0.5 , ],
            [ 0.5 , 0.5 , -0.5 , -0.5 , 0.5 , 0.5 , -0.5 , -0.5 , ],
            [ 0.5 , 0.5 , 0.5 , 0.5 , -0.5 , -0.5 , -0.5 , -0.5 , ],
            [ 1.0 , 1.0 , 1.0 , 1.0 , 1.0 , 1.0 , 1.0 , 1.0 , ]]
          )
  #corners_3d = my_compute_box_3d(boxes_3d[:3],boxes_3d[6],boxes_3d[3:6])
  corners_3d = my_compute_box_3d_jit(corners_3d_in, boxes_3d[:3],boxes_3d[6],boxes_3d[3:6])

  #adjust bbox
  #new_corners_3d = adjust_bbox(boxes_3d,corners_3d)

  head_pnt = (corners_3d[1] + corners_3d[2]) / 2.0
  tail_pnt = (corners_3d[0] + corners_3d[3]) / 2.0
  inter_len = 1.5
  total_len = 6.0
  if car_type == 3:
    inter_len = 1.35
    total_len = 12.42
  if car_type == 5:
    inter_len = 2.25
    total_len = 4.5
  if car_type == 1:
    inter_len = 2.3
    total_len = 4.6
  lidar_loc = (tail_pnt - head_pnt) * inter_len / (1.0*total_len) + head_pnt
  #print(head_pnt, lidar_loc, tail_pnt)

  #Trans = get_T()
  tmp_loc = np.ones([4])
  tmp_loc[:3] = lidar_loc[:3]
  world_loc = np.dot(Trans,tmp_loc)
  #print(world_loc)

  out_BL = Inverse_v2(world_loc[:2])

  #center point transfer
#  Trans = get_T()
  tmp_center_loc = np.ones([4])
  tmp_center_loc[:3] = boxes_3d[:3]
  world_center_loc = np.dot(Trans,tmp_center_loc)

  out_center_BL = Inverse_v2(world_center_loc[:2])
  
  return lidar_loc, out_BL, out_center_BL
