import math
import numpy as np
import wgs84
from rotations import degree2rad, rad2degree
from numba import jit

ZoneOffset = 1e6
ZoneWidth = 3
EastOffset = 500000.0

def Convert_v2(inLonLat): # BL --> XY
    Zone = (int)((inLonLat[1] + 1.5) / ZoneWidth)
    L0 = Zone * ZoneWidth

    dL = inLonLat[1] - L0

    dLRadian = degree2rad(dL)
    latRadian = degree2rad(inLonLat[0])

    X = wgs84.GetX(latRadian)
    N = wgs84.GetN(latRadian)

    t = math.tan(latRadian)
    t2 = t * t
    sinB = math.sin(latRadian)
    cosB = math.cos(latRadian)
    ita2 = wgs84.ep2 * cosB * cosB

    temp = dLRadian * cosB
    temp2 = temp * temp
    yPart1 = 0.5 * N * temp * temp * t
    yPart2 = N * t * (5 - t2 + 9 * ita2 + 4 * ita2 * ita2) * temp2 * temp2 / 24
    yPart3 = temp2 * temp2 * temp2 * N * t * (61 - 58 * t2 + t2 * t2) / 720

    y = X + yPart1 + yPart2 + yPart3

    xPart1 = N * temp
    xPart2 = (1 - t2 + ita2) * N * temp * temp2 / 6
    xPart3 = (5 - 18 * t2 + t2 * t2 + 14 * ita2 - 58 * ita2 * t2) * N * temp * temp2 * temp2 / 120

    x = Zone * ZoneOffset + EastOffset + xPart1 + xPart2 + xPart3

    projPt = np.array([x, y])
    return projPt

@jit
def Inverse_v2(inXY): # XY --> BL

    Zone = (int)(inXY[0] / ZoneOffset)

    L0 = Zone * ZoneWidth
    L0_radian = degree2rad(L0)

    X0 = Zone * ZoneOffset + EastOffset
    Y0 = 0

    x = inXY[0] - X0
    y = inXY[1] - Y0

    Br = wgs84.GetLatByX(y)#


    cosB = math.cos(Br)
    secB = 1 / cosB
    ita2 = wgs84.ep2 * cosB * cosB
    t = math.tan(Br)
    t2 = t * t
    V = wgs84.GetV(Br)
    V2 = V * V
    N = wgs84.c / V
    M = N/V/V
    D = x/N

    tmp3 = (1 + 2 * t2 + ita2) * D * D * D / 6
    tmp5 = (5 + 28 * t2 + 24 * t2 * t2 + 6 * ita2 + 8 * ita2 * t2) * D * D * D * D * D / 120
    l = secB * (D - tmp3 + tmp5)
    L_radian = L0_radian + l
    tmp2 = D * D / 2
    tmp4 = (5 + 3 * t2 + ita2 - 9 * ita2 * t2) * D * D * D * D / 24
    tmp6 = (61 + 90 * t2 + 45 * t2 * t2) * D * D * D * D * D * D / 720
    B_radian = Br - t * V2 * (tmp2 - tmp4 + tmp6)

    B = rad2degree(B_radian)
    L = rad2degree(L_radian)

    outLonLat = np.array([B, L])
    return outLonLat

