#ifndef _YOLOV5_HPP_
#define _YOLOV5_HPP_

#include <chrono>
#include <iostream>
#include <string>
#include <vector>
#include <sys/time.h>

#include "common.hpp"
#include "cuda_runtime_api.h"
#include <omp.h>
#include "config.hpp"



class YoloV5{
    public:
        YoloV5();
        ~YoloV5();

        void Init(std::string enginepath,int batchSize);
        void preprocess(std::vector<cv::Mat>& imgs);
        void infer();
        void post();
        std::map<int,Results*> GetResults();
        

    private:
        int input_w_;
        int input_h_;
        int input_size_;
        int output_size_;
        int batchSize_;
        std::string enginepath_;

        IRuntime* runtime_ = nullptr;
        ICudaEngine* engine_ = nullptr;
        IExecutionContext* context_ = nullptr;
        cudaStream_t stream_;
        void* buffers[2];
        float* inputdata_; // cpu input
        float* outputdata_;

        std::vector<cv::Mat> oriimgs;
        std::map<int,Results*> results_map_;

};

#endif